/*
**	HEI.C - Platform independent code for communicating with Host Automation Products
**			line of ethernet modules.
** 
**	Copyright (C) - 1996-1997 Host Automation Products, Inc.
**
*/

#include "defs.h"
#include <memory.h>
#include <string.h>
#include "hei.h"
#include "function.h"

/* Interface functions */
int HEIIOpen();
int HEIIClose();
int HEIIOpenDevice(HEITransport *pTransport, HEIDevice *pDevice);
int HEIICloseDevice(HEIDevice *pDevice);
int HEIIOpenTransport(HEITransport *pTransport);
int HEIICloseTransport(HEITransport *pTransport);
int HEIIReceivePacket(HEIDevice *pDevice, BYTE *pResponse, int *pResponseSize);
int HEIISendPacket(HEIDevice *pDevice, BYTE *pPacket, WORD PacketSize);
DWORD HEIIGetCounter();

#define DEBUG_FILE 0

#if DEBUG_FILE
#include <stdio.h>
#include <stdarg.h>
#include <windows.h>

FILE *pOutFile=NULL;

void DebugString(char  *String, ...)
	{
	char  Buffer[200];

	va_list ap;

	va_start(ap, String);

	wvsprintf(Buffer, String, ap);

   	/*OutputDebugString(Buffer); */
   	pOutFile = fopen("c:\\heiout.txt", "a+");

	if (pOutFile)
		{
		fputs("\n", pOutFile);
		fputs(Buffer, pOutFile);
		fclose(pOutFile);
		}

	va_end(ap);
	}


#endif


/*
** This macro will return the difference between a start time (in milliseconds) and an end time (also in milliseconds).
** It handles the fact that a DWORD millisecond indicator will wrap every 49.XXX days.
*/
#define TimeDiff(StartTime, EndTime)  (((DWORD)EndTime > (DWORD)StartTime) ? ((DWORD)EndTime - (DWORD)StartTime) : ((DWORD)EndTime + ((DWORD)ULONG_MAX - (DWORD)StartTime) + 1))

int DoCRC = 1;

#define EXTRA_TIME_FOR_SETUP_DATA 10

int DoEncrypt(Encryption *pEncrypt, BYTE *ptr, WORD Num);
   

#if defined(SUPERVISOR)
int PrepareSupervisorPacket(HEIDevice *pDevice, BYTE *pPacket, WORD PacketSize);
#endif /* #if defined(SUPERVISOR) */


WORD CalcCRC(WORD    icrc, BYTE   *icp, WORD    icnt);
  
int InsertCRC(BYTE *pBuffer, WORD Len)
	{
	WORD crc;
	
	if (DoCRC)
		{
		crc = CalcCRC(0, (pBuffer+PACKET_HEADER_SIZE), (WORD) (Len - PACKET_HEADER_SIZE));
		*(pBuffer+5) = (BYTE) (crc & 0xFF);
		*(pBuffer+6) = (BYTE) (crc >> 8);
		}
	else
		{
		*(pBuffer+5) = 0;
		*(pBuffer+6) = 0;
		}
		
	return 0;
	}

BOOL CRCCorrect(BYTE *pBuffer, WORD Len)
	{
	WORD ccrc, rcrc;  /* Calculated & returned CRCs */

	if (!DoCRC)
		return TRUE;
	else if (Len < PACKET_HEADER_SIZE)
		return FALSE;
	else
		{
		ccrc = CalcCRC(0, (pBuffer+PACKET_HEADER_SIZE), (WORD) (Len - PACKET_HEADER_SIZE));
		rcrc = (WORD) *(pBuffer+5) | ((WORD) *(pBuffer+6)<<8);
		return (ccrc == rcrc);
		}
	}


/* int GetResponse(HEIDevice *pDevice, BYTE *pResponse, int *pResponseSize, WORD ExtraTime=0, BOOL ProcessTimeout=TRUE, BOOL CheckAppVal=TRUE)*/
int GetResponse(HEIDevice *pDevice, BYTE *pResponse, int *pResponseSize, WORD ExtraTime, BOOL ProcessTimeout, BOOL CheckAppVal)
	{
	unsigned short ThisAppVal = pDevice->LastAppVal;
	DWORD DeviceTimeout = pDevice->Timeout;

	DWORD Timeout = DeviceTimeout + ExtraTime;
	DWORD StartTime = HEIIGetCounter();

	while (1) 
		{
		int Error = HEIIReceivePacket(pDevice, pResponse, pResponseSize);
		
		if (!Error && (*pResponseSize))
			{
			/* We have a packet to look at. */
			/* Check the App Val. */
			DWORD flag = (WORD) *(pResponse+3) | ((WORD) *(pResponse+4) << 8);
					
			if (flag == ThisAppVal || !ThisAppVal || !CheckAppVal)
				{
				int Retval;
				
				if (!CheckAppVal)
					pDevice->LastAppVal = flag;

				Retval = 0;

				if (CRCCorrect((BYTE *) pResponse, (WORD) *pResponseSize))
					{
					/* We have our response! */
					}
				else 
					{
					pDevice->BadCRCCount++;
					Retval =  HEIE_CRC_MISMATCH;
					}
							
				return Retval;
				}
			else
				{
				pDevice->LatePacketCount++; 
				}
			}
		else if (!ProcessTimeout)
			{
			/* No more packets to look at. */
			return HEIE_NO_RESPONSE;
			}
			
		if (ProcessTimeout && (TimeDiff(StartTime, HEIIGetCounter()) >= Timeout))
			{
			/* Timeout! */
			break;
			}
		}
			
	*pResponseSize = 0;
	return HEIE_TIMEOUT;
	}


static unsigned short AppVal = 1;

__declspec(dllexport) int HEIGetResponse(HEIDevice *pDevice, BYTE *pResponse, int *pResponseSize, BOOL CheckAppVal)
	{
	int Retval;
	
#if DEBUG_FILE
	DebugString("In GetResponse %u", AppVal);
#endif	

	Retval = GetResponse(pDevice, pResponse, pResponseSize, 0, FALSE, CheckAppVal);
	
#if DEBUG_FILE
	DebugString("Leaving GetResponse %u -> %d (0x%X)", AppVal, Retval, Retval);
#endif	

	return Retval;
	}
	

	
/*int SendPacket(HEIDevice *pDevice, BYTE *pPacket, WORD PacketSize, BYTE *pResponse, int *pResponseSize, BOOL WaitForResponse=TRUE, BOOL ReturnWarnings=FALSE)*/
int SendPacket(HEIDevice *pDevice, BYTE *pPacket, WORD PacketSize, BYTE *pResponse, int *pResponseSize, BOOL WaitForResponse, BOOL ReturnWarnings)
	{
	BYTE *ptr;
	unsigned short ThisAppVal;
	BYTE *pBegin = pPacket;
	WORD PreHeaderBytes = pDevice->DataOffset - PACKET_HEADER_SIZE;
	WORD Retry;
	WORD bytes_chksmd;

	if (pDevice->UseAddressedBroadcast)
		PreHeaderBytes -= 7;
			
	pPacket += PreHeaderBytes;
			
	ptr = pPacket;
			
	if (pDevice->UseAddressedBroadcast)
		{
		pPacket[PACKET_HEADER_SIZE] = FUN_ADDRESSED_BROADCAST;
		memcpy(&pPacket[PACKET_HEADER_SIZE+1], pDevice->ENetAddress, 6);
		}
		
	*ptr++ = 'H';
	*ptr++ = 'A';
	*ptr   = 'P';
			
	/* Calculate CRC */
	InsertCRC(pPacket, (WORD) (PacketSize-PreHeaderBytes));
		
	/* Insert the number of bytes which were checksumed. */
	bytes_chksmd = (WORD) PacketSize - PreHeaderBytes - PACKET_HEADER_SIZE;
	*(pPacket+7) = (BYTE) (bytes_chksmd & 0xFF);
	*(pPacket+8) = (BYTE) (bytes_chksmd >> 8);

	/* Do encryption (if required) */
	if (pDevice->_pTransport->Encrypt.Algorithm)
		DoEncrypt(&pDevice->_pTransport->Encrypt, pPacket+5, (WORD) (PacketSize-5-PreHeaderBytes));
		

#if defined(SUPERVISOR)
	PrepareSupervisorPacket(pDevice, pPacket, PacketSize);
#endif	/* #if defined(SUPERVISOR) */

	for (Retry = 0; Retry<=pDevice->Retrys; Retry++)
   		{
		int ResponseSize = *pResponseSize;
		int Error;
				
		ThisAppVal = AppVal++;
		pDevice->LastAppVal = ThisAppVal;
				
		if (!AppVal)
			AppVal++;
		
		if (Retry)
			pDevice->RetryCount++;
					
		/* Insert the app value. */
		*(pPacket+3) = (BYTE) (ThisAppVal & 0xFF);
		*(pPacket+4) = (BYTE) (ThisAppVal >> 8);

		Error = HEIISendPacket(pDevice, pBegin, PacketSize);
				
		if (Error)
			{
			*pResponseSize = 0;
			return Error;
			}
				
		if (!WaitForResponse)
			{
			*pResponseSize = 0;
			return HEIE_NO_RESPONSE;
			}

		Error = GetResponse(pDevice, pResponse, &ResponseSize, 0, TRUE, TRUE);

		if (!Error)
			{
			*pResponseSize = ResponseSize;

			if (Retry != 0 && ReturnWarnings)
				return HEIW_RETRY;

			return 0;   /* Success! */
			}
		else if (Error != HEIE_TIMEOUT)
			return Error;
		}
		
	*pResponseSize = 0;
	return HEIE_TIMEOUT;
	}


/* Transport Stuff */
__declspec(dllexport) int HEIOpenTransport(HEITransport *pTransport, WORD Ver, ENetAddress *pSourceAddress)
/* __declspec(dllexport) int HEIOpenTransport(HEITransport *pTransport, WORD Ver, DWORD NetworkAddress)*/
	{
	int Retval;
	
#if DEBUG_FILE
	DebugString("In HEIOpenTransport");
#endif
	if (HEIAPIVERSION != Ver)
		return HEIE_VER_MISMATCH;

	if (!pTransport)
		return HEIE_NULL_TRANSPORT;
		
	pTransport->Encrypt.Algorithm = HEIEN_NONE;
	/*pTransport->NetworkAddress = NetworkAddress;*/
	pTransport->pSourceAddress = pSourceAddress;
	
	Retval = HEIIOpenTransport(pTransport);
	
#if DEBUG_FILE
	DebugString("Leaving HEIOpenTransport", Retval);
#endif
	
	return Retval;
	}
	
	
__declspec(dllexport) int HEICloseTransport(HEITransport *pTransport)
	{
	if (!pTransport)
		return HEIE_NULL_TRANSPORT;
		
	return HEIICloseTransport(pTransport);
	}

__declspec(dllexport) int HEIOpenDevice
		(
		HEITransport *pTransport, 
		HEIDevice *pDevice, 
		WORD Ver, 
		WORD iTimeout, 
		WORD iRetrys, 
		BOOL UseBroadcast
		)
	{
	int Retval;
	
#if DEBUG_FILE
	DebugString("In HEIOpenDevice");
#endif
	
	if (HEIAPIVERSION != Ver)
		return HEIE_VER_MISMATCH;

	if (!pTransport)
		return HEIE_NULL_TRANSPORT;
		
	if (!pDevice)
		return HEIE_INVALID_DEVICE;

	pDevice->_dwParam				= 0;
	pDevice->Timeout				= iTimeout;
	pDevice->Retrys				= iRetrys;
	pDevice->UseAddressedBroadcast	= UseBroadcast;
	pDevice->UseBroadcast		= UseBroadcast;
	pDevice->pData					= 0;
	pDevice->SizeOfData			= 0;
	pDevice->DataOffset			= PACKET_HEADER_SIZE;
	pDevice->pBuffer				= 0;
	pDevice->RetryCount			= 0;
	pDevice->BadCRCCount			= 0;
	pDevice->LatePacketCount	= 0; 
	pDevice->ParallelPackets	= FALSE;
	pDevice->_pTransport = pTransport;

	if (pDevice->UseAddressedBroadcast)
		pDevice->DataOffset += 7;

	Retval = HEIIOpenDevice(pTransport, pDevice);
	
#if DEBUG_FILE
	DebugString("Leaving HEIOpenDevice", Retval);
#endif
	return Retval;
	}

__declspec(dllexport) int HEICloseDevice(HEIDevice *pDevice)
	{
	int Retval;
	
	if (!pDevice)
		return HEIE_INVALID_DEVICE;

	Retval = HEIICloseDevice(pDevice);
	
	pDevice->_pTransport = (HEITransport *) NULL;
	
	return Retval;
	}


__declspec(dllexport) int HEIOpen(WORD Ver)
	{
	if (HEIAPIVERSION != Ver)
		return HEIE_VER_MISMATCH;
	
	return HEIIOpen();
	}

__declspec(dllexport) int HEIClose()
	{
	return HEIIClose();
	}


int DoEncrypt(Encryption *pEncrypt, BYTE *ptr, WORD Num)
	{
	switch (pEncrypt->Algorithm)
		{
		case 1:		/* Private key encryption (XOR) */
			{
			BYTE *keyptr = pEncrypt->Key;
			WORD x;
			
			for (x=0; x<Num; x++)
				{
				*ptr++ ^= *keyptr++;
				
				if (!*keyptr)
					keyptr = pEncrypt->Key;
				}
				
			return 0;
			}
			
		default:
			return -1;
		}
	}

int OpenQueryableDevice(HEITransport *pTransport, HEIDevice *pDevice, WORD Ver)
	{
	/* Open a device which I can use for broadcasting a packet. */
	int RetVal = HEIOpenDevice(pTransport, pDevice, Ver, 20, 0, TRUE);

	if (!RetVal)
		{
		pDevice->UseAddressedBroadcast = FALSE;    /* Don't want to use 'Addressed' Broadcast, just Broadcast! */
		pDevice->DataOffset -= 7;  /* We are not using addressed broadcast. */
		pDevice->wParam = 0;
		}

	return RetVal;
	}
	
	
int DoQuery(HEITransport *pTransport, HEIDevice *pDevice, HEIDevice *pDevices, WORD *pNumDevices, WORD Ver, BYTE *pBuffer, WORD BufferSize)
	{
	BYTE RetBuffer[600];
	int ResponseSize = sizeof(RetBuffer);
	BYTE FromAddress[20];					
	int Error;
	WORD Count = 0;
	DWORD Timeout = 600;
	DWORD StartTime;
	int Retval;

	pDevice->pData = FromAddress;
	pDevice->SizeOfData = sizeof(FromAddress);

	Error = SendPacket(pDevice, pBuffer, BufferSize, RetBuffer, &ResponseSize, TRUE, FALSE);

	StartTime = HEIIGetCounter();
	Retval = HEIE_NULL;

	while (1) 
		{
		switch ((WORD) Error)
			{
			case HEIE_NULL:
				{
				/* No Error, so process packet. */
				if (RetBuffer[0] == 'H' && RetBuffer[1] == 'A' && (RetBuffer[2] == 'P' || RetBuffer[2] == 'A') &&
					RetBuffer[PACKET_HEADER_SIZE] == 0x55 &&
					(BYTE) RetBuffer[PACKET_HEADER_SIZE+1] == 0xAA)
					{
					if (Count < (*pNumDevices))
						{
						memcpy(pDevices[Count].Address.Raw, pDevice->pData, pDevice->SizeOfData);
						memcpy(pDevices[Count].ENetAddress, (char *) (RetBuffer+PACKET_HEADER_SIZE+2), 6);

						if (pTransport->Protocol == HEIP_IP)
							{
							if (RetBuffer[PACKET_HEADER_SIZE+8] == 1)
								pDevices[Count].Address.Raw[19] = 1; /* The modules IP address is not initialized. */
							else 
								pDevices[Count].Address.Raw[19] = 0;
							}

						Count++;
						}
					else
						{
						Count++;
						Retval = HEIE_BUFFER_TOO_SMALL;
						}
					}

				break;
				}

			case HEIE_TIMEOUT:
				break;
				
			case HEIW_RETRY:
				break;

			default:
				{
				/* Other error, so bail out. */
				pDevice->pData = 0;
				pDevice->SizeOfData = 0;
				return Error;
				}
			}

		if (TimeDiff(StartTime, HEIIGetCounter()) < Timeout)
			{
			ResponseSize = sizeof(RetBuffer);
			pDevice->SizeOfData = sizeof(FromAddress);
			Error = GetResponse(pDevice, RetBuffer, &ResponseSize, 0, TRUE, TRUE);
			}
		else
			break;
		}

	*pNumDevices = Count;
	pDevice->pData = 0;
	pDevice->SizeOfData = 0;
	return Retval;
	}	


__declspec(dllexport) int HEIQueryDevices(HEITransport *pTransport, HEIDevice *pDevices, WORD *pNumDevices, WORD Ver)
	{
	HEIDevice qDevice;
	WORD DataOffset;
	BYTE Buffer[100];
	WORD Total;
	int Retval;
	
	int Error = OpenQueryableDevice(pTransport, &qDevice, Ver);
	
	if (Error)
		return Error;

	/* Build the buffer.		 */
	DataOffset = qDevice.DataOffset;

	Buffer[DataOffset] = FUN_POLLING_ALL;	 	/* This is the function code! */

	Total = DataOffset+1;

	Retval = DoQuery(pTransport, &qDevice, pDevices, pNumDevices, Ver, Buffer, Total);

	HEICloseDevice(&qDevice);
	
	return Retval;	
	}

__declspec(dllexport) int HEIQueryDeviceData(HEITransport *pTransport, HEIDevice *pDevices, WORD *pNumDevices, WORD Ver, WORD DataType, BYTE *pData, WORD SizeofData)
	{
	HEIDevice qDevice;
	WORD DataOffset;
	BYTE Buffer[600];
	WORD Total;
	int Retval;
	
	int Error = OpenQueryableDevice(pTransport, &qDevice, Ver);
	
	if (Error)
		return Error;

	/* Build the buffer.		 */
	DataOffset = qDevice.DataOffset;

	Buffer[DataOffset] = FUN_QUERY_SETUP_DATA;	 	/* This is the function code! */
	Buffer[DataOffset+1] = 0;

	if (SizeofData > 512)
		 SizeofData = 512;

	Buffer[DataOffset+2] = (BYTE) (DataType & 0xFF);
	Buffer[DataOffset+3] = (BYTE) (DataType >> 8);
	Buffer[DataOffset+4] = (BYTE) (SizeofData & 0xFF);
	Buffer[DataOffset+5] = (BYTE) (SizeofData >> 8);

	memcpy(Buffer+DataOffset+6, pData, SizeofData);

	Total = qDevice.DataOffset+6+SizeofData;

	Retval = DoQuery(pTransport, &qDevice, pDevices, pNumDevices, Ver, Buffer, Total);

	HEICloseDevice(&qDevice);
	
	return Retval;	
	}


/* SUPPORT INFORMATION */
__declspec(dllexport) int HEIReadSupportInfo(HEIDevice *pDevice, BYTE *pSupportInfo, WORD SizeOfSupportInfo)
	{
	BYTE Buffer[200];
	BYTE RetBuffer[200];
	WORD DataOffset = pDevice->DataOffset;
	WORD Total;
	int NumBytes;
	int Error;
	int Retval;

	Buffer[DataOffset] = (BYTE) FUN_READ_SUPPORT_INFO;	 		/* This is the function code! */

	Total = DataOffset+1;

	NumBytes = sizeof(RetBuffer);
		
	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &NumBytes, TRUE, FALSE);

	/* Check for error and not warning. */
	if (Error && !(Error & HEIW_FIRST_WARNING))
		return Error;

	Retval = Error;

	if (NumBytes)
		{
		NumBytes -= PACKET_HEADER_SIZE;

		if ((WORD) NumBytes > SizeOfSupportInfo)
			{
			Retval = HEIE_BUFFER_TOO_SMALL;
			NumBytes = SizeOfSupportInfo;
			}

		memcpy(pSupportInfo, RetBuffer+PACKET_HEADER_SIZE, NumBytes);
		}
	else
		{
		Retval = HEIE_ZERO_BYTES_RECEIVED;
		}

	return Retval;
	}


/* VERSION INFORMATION */
__declspec(dllexport) int HEIReadVersionInfo(HEIDevice *pDevice, BYTE *pVerInfo, WORD SizeVerInfo)
	{
	BYTE Buffer[200];
	BYTE RetBuffer[200];
	WORD DataOffset = pDevice->DataOffset;
	WORD Total;
	int Error;
	int Retval;
	int NumBytes;

	Buffer[DataOffset] = (BYTE) FUN_READ_VER_INFO;	 		/* This is the function code! */

	Total = DataOffset+1;

	NumBytes = sizeof(RetBuffer);
		
	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &NumBytes, TRUE, FALSE);

	/* Check for error and not warning. */
	if (Error && !(Error & HEIW_FIRST_WARNING))
		return Error;

	Retval = Error;

	if (NumBytes)
		{
		NumBytes -= PACKET_HEADER_SIZE;

		if ((WORD) NumBytes > SizeVerInfo)
			{
			Retval = HEIE_BUFFER_TOO_SMALL;
			NumBytes = SizeVerInfo;
			}

		memcpy(pVerInfo, RetBuffer+PACKET_HEADER_SIZE, NumBytes);
		}
	else
		{
		Retval = HEIE_ZERO_BYTES_RECEIVED;
		}

	return Retval;
	}


/* BASE DEFINITION */
__declspec(dllexport) int HEIReadBaseDef(HEIDevice *pDevice, BYTE *pBaseDefInfo, WORD *pSizeOfBaseDefInfo)
	{
	BYTE Buffer[300];
	BYTE RetBuffer[600];
	WORD DataOffset = pDevice->DataOffset;
	WORD Total;
	int Error;
	int Retval;
	int NumBytes;

	Buffer[DataOffset] = FUN_READ_BASE_DEF;	 		/* This is the function code! */

	Total = DataOffset+1;

	NumBytes = sizeof(RetBuffer);

	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &NumBytes, TRUE, FALSE);

	/* Check for error and not warning. */
	if (Error && !(Error & HEIW_FIRST_WARNING))
		return Error;

	Retval = Error;

	if (NumBytes)
		{
		NumBytes -= PACKET_HEADER_SIZE;

		if ((WORD) NumBytes > *pSizeOfBaseDefInfo)
			{
			Retval = HEIE_BUFFER_TOO_SMALL;
			NumBytes = *pSizeOfBaseDefInfo;
			}

		memcpy(pBaseDefInfo, RetBuffer+PACKET_HEADER_SIZE, NumBytes);
		(*pSizeOfBaseDefInfo) = NumBytes;
		}
	else
		{
		Retval = HEIE_ZERO_BYTES_RECEIVED;
		(*pSizeOfBaseDefInfo) = 0;
		}

	return Retval;
	}



__declspec(dllexport) int HEIWriteBaseDef(HEIDevice *pDevice, BYTE *pInputBaseDef, WORD SizeOfInputBaseDef, BYTE *pOutputBaseDef, WORD *pSizeOfOutputBaseDef)
	{
	BYTE Buffer[600];
	BYTE RetBuffer[600];
	WORD DataOffset = pDevice->DataOffset;
	WORD Total;
	int Error;
	int Retval;
	int NumBytes;
	WORD *pWord;

	Buffer[DataOffset] = FUN_WRITE_BASE_DEF;	 		/* This is the function code! */
	Buffer[DataOffset+1] = 0;
	
	pWord = (WORD *) &Buffer[DataOffset+2];
	*pWord = SizeOfInputBaseDef;

	Total = DataOffset+4+SizeOfInputBaseDef;

	if (Total > sizeof(Buffer))
		{
		return HEIE_DATA_TOO_LARGE;
		}
	
	memcpy(Buffer+DataOffset+4, pInputBaseDef, SizeOfInputBaseDef);


	NumBytes = sizeof(RetBuffer);

	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &NumBytes, TRUE, FALSE);

	/* Check for error and not warning. */
	if (Error && !(Error & HEIW_FIRST_WARNING))
		return Error;
		
	NumBytes = sizeof(RetBuffer);
	Error = GetResponse(pDevice, RetBuffer, &NumBytes, 30000, TRUE, TRUE);

	/* Check for error and not warning. */
	if (Error && !(Error & HEIW_FIRST_WARNING))
		return Error;
	
	Retval = Error;

	if (NumBytes)
		{
		short int *pInt = (short int *) (RetBuffer+PACKET_HEADER_SIZE);
		
		if (*pInt)
			Retval = *pInt;
		else if (pOutputBaseDef)
			{
			NumBytes -= (PACKET_HEADER_SIZE+2);
	
			if ((WORD) NumBytes > *pSizeOfOutputBaseDef)
				{
				Retval = HEIE_BUFFER_TOO_SMALL;
				NumBytes = *pSizeOfOutputBaseDef;
				}
	
			memcpy(pOutputBaseDef, RetBuffer+PACKET_HEADER_SIZE+2, NumBytes);
			(*pSizeOfOutputBaseDef) = NumBytes;
			}
		}
	else
		{
		Retval = HEIE_ZERO_BYTES_RECEIVED;
		
		if (pOutputBaseDef && pSizeOfOutputBaseDef)
			(*pSizeOfOutputBaseDef) = 0;
		}

	return Retval;
	}

/* DEVICE DEFINITION */
__declspec(dllexport) int HEIReadDeviceDef(HEIDevice *pDevice, BYTE *pDeviceDefInfo, WORD SizeOfDeviceDefInfo)
	{
	BYTE Buffer[200];
	BYTE RetBuffer[200];
	WORD DataOffset = pDevice->DataOffset;
	WORD Total;
	int Error;
	int Retval;
	int NumBytes;

	Buffer[DataOffset] = (BYTE) FUN_READ_DEVICE_INFO;	 		/* This is the function code! */

	Total = DataOffset+1;

	NumBytes = sizeof(RetBuffer);
		
	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &NumBytes, TRUE, FALSE);

	/* Check for error and not warning. */
	if (Error && !(Error & HEIW_FIRST_WARNING))
		return Error;

	Retval = Error;
   
	if (NumBytes)
		{
		NumBytes -= PACKET_HEADER_SIZE;

		if ((WORD) NumBytes > SizeOfDeviceDefInfo)
			{
			Retval = HEIE_BUFFER_TOO_SMALL;
			NumBytes = SizeOfDeviceDefInfo;
			}

		memcpy(pDeviceDefInfo, RetBuffer+PACKET_HEADER_SIZE, NumBytes);
		}
	else
		{
		Retval = HEIE_ZERO_BYTES_RECEIVED;
		}

	return Retval;
	}


/* SETUP */
__declspec(dllexport) int HEIReadSetupData(HEIDevice *pDevice, WORD SetupType, BYTE *pData, WORD *pSizeofData)
	{
	BYTE Buffer[100];
	BYTE RetBuffer[600];
	WORD DataOffset = pDevice->DataOffset;
	WORD Total;
	int Error;
	int Retval;
	int NumBytes;
	short int rv;	/* return value */

	Buffer[DataOffset] = (BYTE) FUN_READ_SETUP_DATA;	 		/* This is the function code! */
	Buffer[DataOffset+1] = 0;
	Buffer[DataOffset+2] = (BYTE) (SetupType & 0xFF);
	Buffer[DataOffset+3] = (BYTE) (SetupType >> 8);

	Total = DataOffset+4;

	NumBytes = sizeof(RetBuffer);

	pDevice->Timeout += EXTRA_TIME_FOR_SETUP_DATA;
		
	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &NumBytes, TRUE, FALSE);

	pDevice->Timeout -= EXTRA_TIME_FOR_SETUP_DATA;

	/* Check for error and not warning. */
	if (Error && !(Error & HEIW_FIRST_WARNING))
		return Error;

	Retval = Error;

	if (NumBytes)
		{
		rv = (short int) RetBuffer[PACKET_HEADER_SIZE]
		   | ((short int) RetBuffer[PACKET_HEADER_SIZE+1]<<8);
		NumBytes -= (PACKET_HEADER_SIZE+2);

		if ((WORD) NumBytes > *pSizeofData)
			{
			Retval = HEIE_BUFFER_TOO_SMALL;
			NumBytes = *pSizeofData;
			}

		if (rv)
			Retval = rv;

		(*pSizeofData) = NumBytes;

		memcpy(pData, RetBuffer+PACKET_HEADER_SIZE+2, NumBytes);
		}
	else
		{
		Retval = HEIE_ZERO_BYTES_RECEIVED;
		}

	return Retval;
	}

__declspec(dllexport) int HEIWriteSetupData(HEIDevice *pDevice, WORD SetupType, BYTE *pData, WORD SizeofData)
	{
	BYTE Buffer[600];
	BYTE RetBuffer[200];
	WORD DataOffset = pDevice->DataOffset;
	WORD *pWord;
	WORD Total;
	int Error;
	int Retval;
	int NumBytes;

	Buffer[DataOffset] = (BYTE) FUN_WRITE_SETUP_DATA;	 		/* This is the function code! */
	Buffer[DataOffset+1] = 0;

	pWord = (WORD *) &Buffer[DataOffset+2];
	*pWord = SetupType;

	Total = DataOffset+4+SizeofData;

	if (Total > sizeof(Buffer))
		return HEIE_DATA_TOO_LARGE;

	memcpy(&Buffer[DataOffset+4], pData, SizeofData);

	NumBytes = sizeof(RetBuffer);
		
	pDevice->Timeout += EXTRA_TIME_FOR_SETUP_DATA;

	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &NumBytes, TRUE, FALSE);

	pDevice->Timeout -= EXTRA_TIME_FOR_SETUP_DATA;

	/* Check for error and not warning. */
	if (Error && !(Error & HEIW_FIRST_WARNING))
		return Error;

	NumBytes = sizeof(RetBuffer);
	Error = GetResponse(pDevice, RetBuffer, &NumBytes, 30000, TRUE, TRUE);

	/* Check for error and not warning. */
	if (Error && !(Error & HEIW_FIRST_WARNING))
		return Error;

	Retval = Error;

	if (NumBytes)
		{
		short int *pInt = (short int *) (RetBuffer+PACKET_HEADER_SIZE);

		if (*pInt)
			Retval = *pInt;
		}
	else
		{
		Retval = HEIE_ZERO_BYTES_RECEIVED;
		}

	return Retval;
	}

__declspec(dllexport) int HEIDeleteSetupData(HEIDevice *pDevice, WORD SetupType)
	{
	BYTE Buffer[200];
	BYTE RetBuffer[200];
	WORD DataOffset = pDevice->DataOffset;
	WORD *pWord;
	WORD Total;
	int Error;
	int Retval;
	int NumBytes;

	Buffer[DataOffset] = (BYTE) FUN_DELETE_SETUP_DATA;	 		/* This is the function code! */
	Buffer[DataOffset+1] = 0;

	pWord = (WORD *) &Buffer[DataOffset+2];
	*pWord = SetupType;

	Total = DataOffset+4;

	NumBytes = sizeof(RetBuffer);
		
	pDevice->Timeout += EXTRA_TIME_FOR_SETUP_DATA;

	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &NumBytes, TRUE, FALSE);

	pDevice->Timeout -= EXTRA_TIME_FOR_SETUP_DATA;

	/* Check for error and not warning. */
	if (Error && !(Error & HEIW_FIRST_WARNING))
		return Error;

	NumBytes = sizeof(RetBuffer);

	Error = GetResponse(pDevice, RetBuffer, &NumBytes, 30000, TRUE, TRUE);

	/* Check for error and not warning. */
	if (Error && !(Error & HEIW_FIRST_WARNING))
		return Error;

	Retval = Error;

	if (NumBytes)
		{
		short int *pInt = (short int *) (RetBuffer+PACKET_HEADER_SIZE);

		if (*pInt)
			Retval = *pInt;
		}
	else
		{
		Retval = HEIE_ZERO_BYTES_RECEIVED;
		}

	return Retval;
	}

__declspec(dllexport) int HEIEnumSetupData(HEIDevice *pDevice, WORD *pData, WORD *pSizeofDataInWords)
	{
	BYTE Buffer[200];
	BYTE RetBuffer[600];
	WORD DataOffset = pDevice->DataOffset;
	WORD Total;
	int Error;
	int Retval;
	int NumBytes;

	Buffer[DataOffset] = (BYTE) FUN_ENUM_SETUP_DATA;	 		/* This is the function code! */

	Total = DataOffset+1;

	NumBytes = sizeof(RetBuffer);

	pDevice->Timeout += EXTRA_TIME_FOR_SETUP_DATA;
	
	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &NumBytes, TRUE, FALSE);

	pDevice->Timeout -= EXTRA_TIME_FOR_SETUP_DATA;

	/* Check for error and not warning. */
	if (Error && !(Error & HEIW_FIRST_WARNING))
		return Error;

	Retval = Error;

	if (NumBytes)
		{
		/*int NumWords; */
		short int *pInt = (short int *) (RetBuffer+PACKET_HEADER_SIZE);
		short int *pNumWords = (short int *) (RetBuffer+PACKET_HEADER_SIZE+2);

		if ((WORD) *pNumWords > *pSizeofDataInWords)
			{
			Retval = HEIE_BUFFER_TOO_SMALL;
			*pNumWords = *pSizeofDataInWords;
			}

		if (*pInt)
			Retval = *pInt;

		(*pSizeofDataInWords) = *pNumWords;

		memcpy((char *) pData, (char *) (RetBuffer+PACKET_HEADER_SIZE+4), *pNumWords * 2);
		}
	else
		{
		Retval = HEIE_ZERO_BYTES_RECEIVED;
		}

	return Retval;
	}



/* IO ACCESS */
__declspec(dllexport) int HEIReadIO(HEIDevice *pDevice, BYTE *pData, WORD *DataSize)
	{
	BYTE Buffer[100];
	BYTE RetBuffer[1200];
	WORD DataOffset = pDevice->DataOffset;
	WORD Total;
	int Error;
	int Retval;
	int NumBytes;

	Buffer[DataOffset] = FUN_READ_IO;	 	/* This is the function code! */

	Total = DataOffset+1;

	NumBytes = sizeof(RetBuffer);
	/*NumBytes = *DataSize; */

	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &NumBytes, (pDevice->ParallelPackets) ? FALSE : TRUE, TRUE);

	/* Check for error and not warning. */
	if (Error && !(Error & HEIW_FIRST_WARNING))
		return Error;

	Retval = Error;

	if (NumBytes)
		{
		short int *pInt = (short int *) &RetBuffer[PACKET_HEADER_SIZE];

		NumBytes -= PACKET_HEADER_SIZE;
		NumBytes -= 2;

		if ((WORD) NumBytes > *DataSize)
			{
			Retval = HEIE_BUFFER_TOO_SMALL;
			NumBytes = *DataSize;
			}

		if (*pInt)
			Retval = (*pInt) | 0x8000;

		memcpy(pData, RetBuffer+PACKET_HEADER_SIZE+2, NumBytes);
		(*DataSize) = NumBytes;
		}
	else
		{
		Retval = HEIE_ZERO_BYTES_RECEIVED;
		(*DataSize) = 0;
		}

	return Retval;
	}


__declspec(dllexport) int HEIWriteIO(HEIDevice *pDevice, BYTE *pData, WORD SizeofData, BYTE *pReturnData, WORD *pSizeofReturnData)
	{
	BYTE Buffer[1200];
	BYTE RetBuffer[1200];
	WORD DataOffset = pDevice->DataOffset;
	WORD Total;
	int Error;
	int Retval;
	int NumBytes;

	Buffer[DataOffset] = FUN_WRITE_IO;	 		/* This is the function code! */
	
	Total = DataOffset+1+SizeofData;

	if (Total > sizeof(Buffer))
		return HEIE_DATA_TOO_LARGE;

	memcpy(Buffer+DataOffset+1, pData, SizeofData);

	NumBytes = sizeof(RetBuffer);
	/*NumBytes = *pSizeofReturnData; */
		
	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &NumBytes, (pDevice->ParallelPackets) ? FALSE : TRUE, TRUE);

	/* Check for error and not warning. */
	if (Error && !(Error & HEIW_FIRST_WARNING))
		return Error;

	Retval = Error;

	if (NumBytes)
		{
		short int *pInt = (short int *) &RetBuffer[PACKET_HEADER_SIZE];

		if (pReturnData && pSizeofReturnData)
			{
			NumBytes -= PACKET_HEADER_SIZE;
			NumBytes -= 2;

			if ((WORD) NumBytes > *pSizeofReturnData)
				{
				Retval = HEIE_BUFFER_TOO_SMALL;
				NumBytes = *pSizeofReturnData;
				}

			memcpy(pReturnData, RetBuffer+PACKET_HEADER_SIZE+2, NumBytes);
			(*pSizeofReturnData) = NumBytes;
			}

		if (*pInt)
			Retval = (*pInt) | 0x8000;
		}
	else
		{
		Retval = HEIE_ZERO_BYTES_RECEIVED;

		if (pSizeofReturnData)
			(*pSizeofReturnData) = 0;
		}

	return Retval;
	}



__declspec(dllexport) int HEIPetDevice(HEIDevice *pDevice)
	{
	BYTE Buffer[200];
	BYTE RetBuffer[200];
	WORD DataOffset = pDevice->DataOffset;
	WORD Total;
	int NumBytes;

	Buffer[DataOffset] = FUN_PET_LINK;

	Total = DataOffset+1;

	NumBytes = sizeof(RetBuffer);

	return SendPacket(pDevice, Buffer, Total, RetBuffer, &NumBytes, TRUE, FALSE);
	}

__declspec(dllexport) int HEIReadEthernetStats(HEIDevice *pDevice, BYTE *pData, WORD *DataSize, BOOL Clear)
	{
	BYTE Buffer[200];
	BYTE RetBuffer[200];
	WORD DataOffset = pDevice->DataOffset;
	WORD Total;
	int Error;
	int Retval;
	int NumBytes;

	Buffer[DataOffset] = FUN_READ_ETHERNET_STATS;	 	/* This is the function code! */
	Buffer[DataOffset+1] = 0;
	Buffer[DataOffset+2] = Clear ? 1 : 0;

	Total = DataOffset+3;

	NumBytes = sizeof(RetBuffer);
		
	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &NumBytes, TRUE, FALSE);

	/* Check for error and not warning. */
	if (Error && !(Error & HEIW_FIRST_WARNING))
		return Error;

	Retval = Error;

	if (NumBytes)
		{
		NumBytes -= PACKET_HEADER_SIZE;

		if ((WORD) NumBytes > *DataSize)
			{
			Retval = HEIE_BUFFER_TOO_SMALL;
			NumBytes = *DataSize;
			}

		memcpy(pData, RetBuffer+PACKET_HEADER_SIZE, NumBytes);
		(*DataSize) = NumBytes;
		}
	else
		{
		Retval = HEIE_ZERO_BYTES_RECEIVED;
		(*DataSize) = 0;
		}

	return Retval;
	}


__declspec(dllexport) int HEIReadModuleStatus(HEIDevice *pDevice, BYTE *pData, WORD *DataSize, BOOL Reset)
	{
	BYTE Buffer[200];
	BYTE RetBuffer[600];
	WORD DataOffset = pDevice->DataOffset;
	WORD Total;
	int Error;
	int Retval;
	int NumBytes;

	Buffer[DataOffset] = FUN_READ_MODULE_STATUS;	 	/* This is the function code! */
	Buffer[DataOffset+1] = 0;
	Buffer[DataOffset+2] = Reset ? 1 : 0;

	Total = DataOffset+3;

	NumBytes = sizeof(RetBuffer);
		
	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &NumBytes, TRUE, FALSE);

	/* Check for error and not warning. */
	if (Error && !(Error & HEIW_FIRST_WARNING))
		return Error;

	Retval = Error;

	if (NumBytes)
		{
		short int *pInt = (short int *) &RetBuffer[PACKET_HEADER_SIZE];

		NumBytes -= PACKET_HEADER_SIZE;
		NumBytes -= 2;

		if ((WORD) NumBytes > *DataSize)
			{
			Retval = HEIE_BUFFER_TOO_SMALL;
			NumBytes = *DataSize;
			}

		if (*pInt)
			Retval = *pInt;

		memcpy(pData, RetBuffer+PACKET_HEADER_SIZE+2, NumBytes);
		(*DataSize) = NumBytes;
		}
	else
		{
		Retval = HEIE_ZERO_BYTES_RECEIVED;
		(*DataSize) = 0;
		}

	return Retval;
	}
	
	
/* CCM STUFF */
#define CCM_READ			0x1E
#define CCM_WRITE			0x20
	
__declspec(dllexport) int HEICCMRequest(HEIDevice *pDevice, BOOL bWrite, BYTE DataType, WORD Address, WORD DataLen, BYTE *pData)
	{
	BYTE Buffer[320];
	BYTE RetBuffer[320];
	WORD DataOffset = pDevice->DataOffset;
	WORD Total;
	int Error;
	int NumBytes;
	BYTE *pRet;
	short int sError;
	
#if DEBUG_FILE
	DebugString("In HEICCMRequest %u", AppVal);
#endif		

	if(DataLen > 256)
		return HEIE_INVALID_REQUEST;
   
   /* Function */
	Buffer[DataOffset] = FUN_CCM_REQUEST;
	
	/* Unused */
	Buffer[DataOffset+1] = 0;

	/* Request Code. */
	Buffer[DataOffset+2] = 0x01;
	
	/* Direction */
	Buffer[DataOffset+3] = bWrite ? CCM_WRITE : CCM_READ;
	
	/* Length */
	Buffer[DataOffset+4] = (DataLen == 256) ? 0 : (BYTE)DataLen;
	
	/* LSB of Address */
	Buffer[DataOffset+5] = (BYTE) (Address & 0xFF);
	
	/* MSB of Address */
	Buffer[DataOffset+6] = (BYTE) ((Address>>8) & 0xFF);
	
	/* Data Type */
	Buffer[DataOffset+7] = DataType;

	Total = DataOffset+8;

	if (bWrite)	
		{
		/* We are writing. */
		memcpy(&Buffer[DataOffset+8], pData, DataLen);
		Total += DataLen;
		}

	NumBytes = sizeof(RetBuffer);

	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &NumBytes, (pDevice->ParallelPackets) ? FALSE : TRUE, TRUE);

	if (Error && !(Error & HEIW_FIRST_WARNING))
		{
#if DEBUG_FILE
		DebugString("Error %d (0x%X) from HEICCMRequest %u", Error, Error, AppVal);
#endif		
		return Error;
		}

	/* Check and make sure we got an ack. */
	pRet = (RetBuffer+PACKET_HEADER_SIZE);
	
	if (*pRet != FUN_ACK || (*(pRet+1) != FUN_CCM_REQUEST))
		{
		sError = (short int) RetBuffer[PACKET_HEADER_SIZE+2]
			| ((short int) RetBuffer[PACKET_HEADER_SIZE+3] << 8);
#if DEBUG_FILE
		DebugString("Error %d (0x%X) from HEICCMRequest %u", sError, sError, AppVal);
#endif		
		return sError;
		}

	NumBytes = sizeof(RetBuffer);
	Error = GetResponse(pDevice, RetBuffer, &NumBytes, 0, TRUE, TRUE);

	if (Error && !(Error & HEIW_FIRST_WARNING))
		{
#if DEBUG_FILE
		DebugString("Error %d (0x%X) from HEICCMRequest %u", Error, Error, AppVal);
#endif		
		return Error;
		}

	NumBytes -= PACKET_HEADER_SIZE;

	pRet = (RetBuffer + PACKET_HEADER_SIZE);

	if (*pRet != FUN_RESPONSE || *(pRet+1) != FUN_CCM_REQUEST)
		{
#if DEBUG_FILE
		Error = HEIE_INVALID_RESPONSE;
		DebugString("Error %d (0x%X) from HEICCMRequest %u", Error, Error, AppVal);
#endif		
		return HEIE_INVALID_RESPONSE;
		}
	
	sError = (short int) RetBuffer[PACKET_HEADER_SIZE+2]
		+ ((short int) RetBuffer[PACKET_HEADER_SIZE+3] << 8);

	if (sError)
		{
#if DEBUG_FILE
		DebugString("Error %d (0x%X) from HEICCMRequest %u", sError, sError, AppVal);
#endif		
		return sError;
		}
		
	NumBytes -= 4;

	if(!bWrite && NumBytes != (int)DataLen)
		{
#if DEBUG_FILE
		Error = HEIE_INVALID_REQUEST;
		DebugString("Error %d (0x%X) from HEICCMRequest %u", Error, Error, AppVal);
#endif		
		return HEIE_INVALID_REQUEST;
		}
	
	memcpy(pData, RetBuffer+PACKET_HEADER_SIZE+4, DataLen);

#if DEBUG_FILE
	DebugString("Leaving HEICCMRequest %u", AppVal);
#endif		
	
	return HEIE_NULL;
	}


__declspec(dllexport) int HEIKSEQRequest(HEIDevice *pDevice, WORD DataLenIn, BYTE *pData, WORD *pDataLen)
	{
	BYTE Buffer[320];
	BYTE RetBuffer[320];
	WORD DataOffset = pDevice->DataOffset;
	WORD Total;
	int Error;
	int NumBytes;
	BYTE *pRet;
	short int *pError;
	int Retval;

#if DEBUG_FILE
	DebugString("In HEIKSEQRequest %u", AppVal);
#endif		

	if(DataLenIn > 256)
		return HEIE_INVALID_REQUEST;

	/* Function */
	Buffer[DataOffset] = FUN_KSEQ_REQUEST;
	
	/* Unused */
	Buffer[DataOffset+1] = 0;

	/* Request Code. */
	Buffer[DataOffset+2] = (DataLenIn == 256) ? 0 : (BYTE)DataLenIn;
	
	/* Unused */
	Buffer[DataOffset+3] = 0;
	
	/* Copy task code data. */
	memcpy(&Buffer[DataOffset+4], pData, DataLenIn);

	Total = DataOffset+4+DataLenIn;

	NumBytes = sizeof(RetBuffer);

	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &NumBytes, (pDevice->ParallelPackets) ? FALSE : TRUE, TRUE);

	if (Error && !(Error & HEIW_FIRST_WARNING))
		return Error;

	/* Check and make sure we got an ack. */
	pRet = (RetBuffer+PACKET_HEADER_SIZE);
	
	if (*pRet != FUN_ACK || (*(pRet+1) != FUN_KSEQ_REQUEST))
		{
		short int *pError = (short int *) (RetBuffer+PACKET_HEADER_SIZE+2);
		
		return *pError;
		}

	NumBytes = sizeof(RetBuffer);
	Error = GetResponse(pDevice, RetBuffer, &NumBytes, 0, TRUE, TRUE);

	if (Error)
		return Error;
		
	NumBytes -= PACKET_HEADER_SIZE;

	pRet = (RetBuffer + PACKET_HEADER_SIZE);

	if (*pRet != FUN_RESPONSE || *(pRet+1) != FUN_KSEQ_REQUEST)
		{
		return HEIE_INVALID_RESPONSE;
		}
	
	pError = (short int *) (RetBuffer+PACKET_HEADER_SIZE+2);
	
	if (*pError)
		return *pError;
		
	NumBytes -= 4;
	
	 Retval = HEIE_NULL;

	if ((WORD) NumBytes > *pDataLen)
		{
		Retval = HEIE_BUFFER_TOO_SMALL;
		NumBytes = *pDataLen;
		}
	
	memcpy(pData, RetBuffer+PACKET_HEADER_SIZE+4, NumBytes);
	*pDataLen = NumBytes;
	
#if DEBUG_FILE
	DebugString("Leaving HEIKSEQRequest", Retval);
#endif		
	
	return Retval;
	}

/* Returns number of bytes not written. */
__declspec(dllexport) int HEIWriteComm(HEIDevice *pDevice, WORD Num2Write, BYTE *pData)
	{
	BYTE Buffer[200];
	BYTE RetBuffer[200];
	WORD DataOffset = pDevice->DataOffset;
	WORD Total;
	WORD *pWord;
	int Error;
	int NumBytes;

	Buffer[DataOffset] = FUN_SERIAL_PORT;
	Buffer[DataOffset+1] = SPF_WRITE_COMM;
	pWord = (WORD *) &Buffer[DataOffset+2];
	*pWord = Num2Write;

	if (Num2Write > (sizeof(Buffer) - DataOffset - 4))
		return HEIE_DATA_TOO_LARGE;

	memcpy(Buffer+DataOffset+4, pData, Num2Write);

	Total = DataOffset+4+Num2Write;

	NumBytes = sizeof(RetBuffer);
		
	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &NumBytes, TRUE, FALSE);

	if (Error) return Error;

	if (NumBytes)
		return *((short int *) (RetBuffer+PACKET_HEADER_SIZE));
	else
		return HEIE_ZERO_BYTES_RECEIVED;
	}

__declspec(dllexport) int HEIReadComm(HEIDevice *pDevice, WORD *pNum2Read, BYTE *pData)
	{
	BYTE Buffer[200];
	BYTE RetBuffer[200];
	WORD DataOffset = pDevice->DataOffset;
	WORD Total;
	WORD *pWord;
	int Error;
	int Retval;
	int NumBytes;

	Buffer[DataOffset] = FUN_SERIAL_PORT;
	Buffer[DataOffset+1] = SPF_READ_COMM;
	pWord = (WORD *) &Buffer[DataOffset+2];
	*pWord = *pNum2Read;

	Total = DataOffset+4;

	NumBytes = sizeof(RetBuffer);
		
	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &NumBytes, TRUE, FALSE);

	if (Error) return Error;

	Retval = Error;

	if (NumBytes)
		{
		short int *pInt = (short int *) (RetBuffer+PACKET_HEADER_SIZE);
		short int *pNumReturned = (short int *) (RetBuffer+PACKET_HEADER_SIZE+2);

		if (*pInt)
			Retval = *pInt;

		(*pNum2Read) = *pNumReturned;

		memcpy(pData, RetBuffer+PACKET_HEADER_SIZE+4, *pNumReturned);
		}
	else
		Retval = HEIE_ZERO_BYTES_RECEIVED;

	return Retval;
	}

__declspec(dllexport) int HEIGetRXAvailable(HEIDevice *pDevice, WORD *pAvailable)
	{
	BYTE Buffer[200];
	BYTE RetBuffer[200];
	WORD DataOffset = pDevice->DataOffset;
	WORD Total;
	int Error;
	int NumBytes;

	Buffer[DataOffset] = FUN_SERIAL_PORT;
	Buffer[DataOffset+1] = SPF_RX_AVAILABLE;

	Total = DataOffset+2;

	NumBytes = sizeof(RetBuffer);
		
	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &NumBytes, TRUE, FALSE);

	if (Error) return Error;

	if (NumBytes)
		{
		*pAvailable = *((WORD *) (RetBuffer+PACKET_HEADER_SIZE+2));
		return *((short int *) (RetBuffer+PACKET_HEADER_SIZE));
		}
	else
		return HEIE_ZERO_BYTES_RECEIVED;
	}

__declspec(dllexport) int HEIFlushRXQueue(HEIDevice *pDevice)
	{
	BYTE Buffer[200];
	BYTE RetBuffer[200];
	WORD DataOffset = pDevice->DataOffset;
	WORD Total;
	int Error;
	int NumBytes;

	Buffer[DataOffset] = FUN_SERIAL_PORT;
	Buffer[DataOffset+1] = SPF_RX_FLUSH;

	Total = DataOffset+2;

	NumBytes = sizeof(RetBuffer);
		
	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &NumBytes, TRUE, FALSE);

	if (Error) return Error;

	if (NumBytes)
		return *((short int *) (RetBuffer+PACKET_HEADER_SIZE));
	else
		return HEIE_ZERO_BYTES_RECEIVED;
	}

__declspec(dllexport) int HEISetupSerialPort(HEIDevice *pDevice, SerialSetup *pSetup, BOOL WriteToFlash)
	{
	BYTE Buffer[200];
	BYTE RetBuffer[200];
	WORD DataOffset = pDevice->DataOffset;
	WORD Total;
	int Error;
	int NumBytes;

	Buffer[DataOffset] = FUN_SERIAL_PORT;
	Buffer[DataOffset+1] = SPF_SETUP;
	Buffer[DataOffset+2] = (BYTE) WriteToFlash;
	Buffer[DataOffset+3] = 0;

	memcpy(Buffer+DataOffset+4, pSetup, sizeof(SerialSetup));

	Total = DataOffset+4+sizeof(SerialSetup);

	NumBytes = sizeof(RetBuffer);

	if (WriteToFlash)
		pDevice->Timeout += EXTRA_TIME_FOR_SETUP_DATA;
		
	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &NumBytes, TRUE, FALSE);

	if (WriteToFlash)
		pDevice->Timeout -= EXTRA_TIME_FOR_SETUP_DATA;

	if (Error) return Error;

	if (WriteToFlash)
		{
		NumBytes = sizeof(RetBuffer);
		Error = GetResponse(pDevice, RetBuffer, &NumBytes, 30000, TRUE, TRUE);
		}

	if (NumBytes)
		return *((short int *) (RetBuffer+PACKET_HEADER_SIZE));
	else
		return HEIE_ZERO_BYTES_RECEIVED;
	}
	
__declspec(dllexport) int HEIReadSerialPortSetup(HEIDevice *pDevice, SerialSetup *pSetup)
	{
	BYTE Buffer[200];
	BYTE RetBuffer[200];
	WORD DataOffset = pDevice->DataOffset;
	WORD Total;
	int Error;
	int NumBytes;

	Buffer[DataOffset] = FUN_SERIAL_PORT;
	Buffer[DataOffset+1] = SPF_READ_SETUP;
	Buffer[DataOffset+2] = 0;
	Buffer[DataOffset+3] = 0;

	Total = DataOffset+4;

	NumBytes = sizeof(RetBuffer);
		
	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &NumBytes, TRUE, FALSE);

	if (Error) return Error;

	if (NumBytes)
		{
		/* Copy data into user buffer. */
		memcpy(pSetup, RetBuffer+PACKET_HEADER_SIZE, sizeof(SerialSetup));
		return 0;
		}
	else
		return HEIE_ZERO_BYTES_RECEIVED;
	}
	

__declspec(dllexport) int HEIGetTXLeft(HEIDevice *pDevice, WORD *pLeft)
	{
	BYTE Buffer[200];
	BYTE RetBuffer[200];
	WORD DataOffset = pDevice->DataOffset;
	WORD Total;
	int Error;
	int Retval;
	int NumBytes;

	Buffer[DataOffset] = FUN_SERIAL_PORT;
	Buffer[DataOffset+1] = SPF_TX_LEFT;

	Total = DataOffset+2;

	NumBytes = sizeof(RetBuffer);
		
	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &NumBytes, TRUE, FALSE);

	if (Error) return Error;

	Retval = Error;

	if (NumBytes)
		{
		*pLeft = *((WORD *) (RetBuffer+PACKET_HEADER_SIZE+2));
		return *((short int *) (RetBuffer+PACKET_HEADER_SIZE));
		}
	else
		return HEIE_ZERO_BYTES_RECEIVED;
	}


/* PLC MEMORY STUFF */
__declspec(dllexport) int HEIReadMemory(HEIDevice *pDevice, WORD Type, DWORD Offset, WORD NumBytes, BYTE *pBuffer)
	{
	BYTE Buffer[200];
	BYTE RetBuffer[600];
	WORD DataOffset = pDevice->DataOffset;
	WORD Total;
	int Error;
	int Num;
	WORD *pWord;
	DWORD *pDWord;

	Buffer[DataOffset] = FUN_READ_MEMORY;
	Buffer[DataOffset+1] = 0;
	
	pWord = (WORD *) &Buffer[DataOffset+2];
	*pWord++ = Type;
	*pWord++ = NumBytes;
	pDWord = (DWORD *) &Buffer[DataOffset+6];
	*pDWord++ = Offset;

	Total = DataOffset+10;

	Num = sizeof(RetBuffer);
		
	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &Num, TRUE, FALSE);

	if (Error) 
		return Error;

	if (Num)
		{
		short int *pError = (short int *) (RetBuffer+PACKET_HEADER_SIZE);
		
		if (!*pError || (*pError == LENGTH_WARNING))
			{
			/* Copy the memory. */
			memcpy(pBuffer, RetBuffer+PACKET_HEADER_SIZE+2, NumBytes);
			}
		
		return *pError;
		}
	else
		return HEIE_ZERO_BYTES_RECEIVED;
		
	return 0;
	}
	
__declspec(dllexport) int HEIWriteMemory(HEIDevice *pDevice, WORD Type, DWORD Offset, WORD NumBytes, BYTE *pBuffer)
	{
	BYTE Buffer[600];
	BYTE RetBuffer[200];
	WORD DataOffset = pDevice->DataOffset;
	WORD Total;
	int Error;
	int Num;
	WORD *pWord;
	DWORD *pDWord;

	if (NumBytes > 512)
		return HEIE_DATA_TOO_LARGE;

	Buffer[DataOffset] = FUN_WRITE_MEMORY;
	Buffer[DataOffset+1] = 0;
	
	pWord = (WORD *) &Buffer[DataOffset+2];
	*pWord++ = Type;
	*pWord++ = NumBytes;
	pDWord = (DWORD *) &Buffer[DataOffset+6];
	*pDWord++ = Offset;
	
	memcpy(Buffer+DataOffset+10, pBuffer, NumBytes);

	Total = DataOffset+10+NumBytes;

	Num = sizeof(RetBuffer);
		
	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &Num, TRUE, FALSE);

	if (Error) return Error;

	if (Num)
		return *((short int *) (RetBuffer+PACKET_HEADER_SIZE));
	else
		return HEIE_ZERO_BYTES_RECEIVED;
		
	return 0;
	}
	

__declspec(dllexport) int HEIAccessMemory(HEIDevice *pDevice, MemRef MemRefs[], WORD NumRefs)
	{
	WORD DataOffset = pDevice->DataOffset;
	BYTE Buffer[600];
	BYTE RetBuffer[600];
	BYTE *pSendData;
	WORD x;
	WORD *pWord;
	WORD ReadSize = 2 + 2;					/* Returned error value is 2 bytes long */
											/* Returned num MemRefs is 2 bytes long */
	WORD TotalSize = 4 + DataOffset;		/* Header info is 4 bytes long */
	int Num;
	int Error;

	Buffer[DataOffset] = FUN_ACCESS_MEMORY;
	Buffer[DataOffset+1] = 0;
	pWord = (WORD *) &Buffer[DataOffset+2];
	*pWord = NumRefs;

	pSendData = &Buffer[DataOffset+4];

	for (x=0; x<NumRefs; x++)
		{
		memcpy(pSendData, &MemRefs[x], sizeof(MemRefDetail));
		TotalSize += sizeof(MemRefDetail);
		pSendData += sizeof(MemRefDetail);

		if (MemRefs[x].Detail.Direction == ACCESS_READ)
			{
			/* Read request. */
			ReadSize += (MemRefs[x].Detail.NumBytes + sizeof(MemRefDetail));
			}
		else if (MemRefs[x].Detail.Direction == ACCESS_WRITE)
			{
			/* Write request. */
			memcpy(pSendData, MemRefs[x].pBuffer, MemRefs[x].Detail.NumBytes);
			TotalSize += MemRefs[x].Detail.NumBytes;
			pSendData += MemRefs[x].Detail.NumBytes;
			}
		else
			{
			return HEIE_INVALID_REQUEST;
			}

		if ((TotalSize > sizeof(Buffer)) || (ReadSize > sizeof(RetBuffer)))
			{
			return HEIE_DATA_TOO_LARGE;
			}
		}

	/* Ready to send. */
	Num = sizeof(RetBuffer);
		
	Error = SendPacket(pDevice, Buffer, TotalSize, RetBuffer, &Num, TRUE, FALSE);

	if (Error) return Error;

	if (Num)
		{
		short int *pError = (short int *) (RetBuffer+PACKET_HEADER_SIZE);
		short int *pNum = (short int *) (RetBuffer+PACKET_HEADER_SIZE+2);
		
		if (!*pError && *pNum)
			{
			/* No errors, and we have some read requests returned. */
			WORD x;
			MemRefDetail *pReturnDetail = (MemRefDetail *) (RetBuffer+PACKET_HEADER_SIZE+4);

			for (x=0; x<NumRefs; x++)
				{
				/* Find a read request. */
				if (MemRefs[x].Detail.Direction == ACCESS_READ)
					{
					/* Read request; Make sure it matches. */
					if (MemRefs[x].Detail.Type		== pReturnDetail->Type && 
						MemRefs[x].Detail.Offset	== pReturnDetail->Offset &&
						MemRefs[x].Detail.NumBytes	== pReturnDetail->NumBytes)
						{
						/* Requests match, so copy data. */
						BYTE *pData = (BYTE *) pReturnDetail;
						pData += sizeof(MemRefDetail);

						memcpy(MemRefs[x].pBuffer, pData, MemRefs[x].Detail.NumBytes);

						/* Reset for next one. */
						pData += MemRefs[x].Detail.NumBytes;
						pReturnDetail = (MemRefDetail *) pData;
						}
					else
						{
						/* Requests don't match, so return error. */
						return HEIE_INVALID_RESPONSE;
						}
					}
				}
			}
		
		return *pError;
		}
	else
		return HEIE_ZERO_BYTES_RECEIVED;
	}


__declspec(dllexport) int HEIENumMemory(HEIDevice *pDevice, WORD *pSizeofDataInTypes, MemoryTypeDef *pData)
	{
	BYTE Buffer[200];
	BYTE RetBuffer[600];
	WORD DataOffset = pDevice->DataOffset;
	WORD Total;
	int Error;
	int Retval;
	int NumBytes;

	Buffer[DataOffset] = (BYTE) FUN_ENUM_MEMORY;	 		/* This is the function code! */

	Total = DataOffset+1;

	NumBytes = sizeof(RetBuffer);

	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &NumBytes, TRUE, FALSE);

	/* Check for error and not warning. */
	if (Error)
		return Error;

	Retval = Error;

	if (NumBytes)
		{
		/*int NumTypes; */
		short int *pInt = (short int *) (RetBuffer+PACKET_HEADER_SIZE);
		short int *pNumTypes = (short int *) (RetBuffer+PACKET_HEADER_SIZE+2);

		if ((WORD) *pNumTypes > *pSizeofDataInTypes)
			{
			Retval = HEIE_BUFFER_TOO_SMALL;
			*pNumTypes = *pSizeofDataInTypes;
			}

		if (*pInt)
			Retval = *pInt;

		(*pSizeofDataInTypes) = *pNumTypes;

		memcpy((char *) pData, (char *) (RetBuffer+PACKET_HEADER_SIZE+4), (*pNumTypes) * sizeof(MemoryTypeDef));
		}
	else
		{
		Retval = HEIE_ZERO_BYTES_RECEIVED;
		}

	return Retval;
	}


__declspec(dllexport) int HEIReadSharedRAM(HEIDevice *pDevice, WORD Base, WORD Slot, WORD Address, WORD Bytes2Read, BYTE *pBuffer)
	{
	BYTE Buffer[200];
	BYTE RetBuffer[600];
	WORD DataOffset = pDevice->DataOffset;
	WORD Total;
	int Error;
	int Num;
	WORD *pWord;

	Buffer[DataOffset] = FUN_READ_SHARED_RAM;
	Buffer[DataOffset+1] = 0;
	
	pWord = (WORD *) &Buffer[DataOffset+2];
	*pWord++ = Base;
	*pWord++ = Slot;
	*pWord++ = Address;
	*pWord++ = Bytes2Read;

	Total = DataOffset+10;

	Num = sizeof(RetBuffer);
		
	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &Num, TRUE, FALSE);

	if (Error) 
		return Error;

	if (Num)
		{
		short int *pError = (short int *) (RetBuffer+PACKET_HEADER_SIZE);
		
		if (!*pError || (*pError == LENGTH_WARNING))
			{
			/* Copy the memory. */
			memcpy(pBuffer, RetBuffer+PACKET_HEADER_SIZE+2, Bytes2Read);
			}
		
		return *pError;
		}
	else
		return HEIE_ZERO_BYTES_RECEIVED;
		
	return 0;
	}
	

__declspec(dllexport) int HEIWriteSharedRAM(HEIDevice *pDevice, WORD Base, WORD Slot, WORD Address, WORD Bytes2Write, BYTE *pBuffer)
	{
	BYTE Buffer[600];
	BYTE RetBuffer[200];
	WORD DataOffset = pDevice->DataOffset;
	WORD Total;
	int Error;
	int Num;
	WORD *pWord;

	if (Bytes2Write > 256)
		return HEIE_DATA_TOO_LARGE;

	Buffer[DataOffset] = FUN_WRITE_SHARED_RAM;
	Buffer[DataOffset+1] = 0;
	
	pWord = (WORD *) &Buffer[DataOffset+2];
	*pWord++ = Base;
	*pWord++ = Slot;
	*pWord++ = Address;
	*pWord++ = Bytes2Write;
	
	memcpy(Buffer+DataOffset+10, pBuffer, Bytes2Write);

	Total = DataOffset+10+Bytes2Write;

	Num = sizeof(RetBuffer);
		
	Error = SendPacket(pDevice, Buffer, Total, RetBuffer, &Num, TRUE, FALSE);

	if (Error) return Error;

	if (Num)
		return *((short int *) (RetBuffer+PACKET_HEADER_SIZE));
	else
		return HEIE_ZERO_BYTES_RECEIVED;
		
	return 0;
	}


/*
** Calculate, 16 Bit CRC
*/
/* number of bits in CRC: don't change it.  */
#define W 16

/* this the number of bits per char: don't change it. */
#define B 8

static unsigned short crctab[1<<B] = { 
    0x0000,  0x1021,  0x2042,  0x3063,  0x4084,  0x50a5,  0x60c6,  0x70e7,
    0x8108,  0x9129,  0xa14a,  0xb16b,  0xc18c,  0xd1ad,  0xe1ce,  0xf1ef,
    0x1231,  0x0210,  0x3273,  0x2252,  0x52b5,  0x4294,  0x72f7,  0x62d6,
    0x9339,  0x8318,  0xb37b,  0xa35a,  0xd3bd,  0xc39c,  0xf3ff,  0xe3de,
    0x2462,  0x3443,  0x0420,  0x1401,  0x64e6,  0x74c7,  0x44a4,  0x5485,
    0xa56a,  0xb54b,  0x8528,  0x9509,  0xe5ee,  0xf5cf,  0xc5ac,  0xd58d,
    0x3653,  0x2672,  0x1611,  0x0630,  0x76d7,  0x66f6,  0x5695,  0x46b4,
    0xb75b,  0xa77a,  0x9719,  0x8738,  0xf7df,  0xe7fe,  0xd79d,  0xc7bc,
    0x48c4,  0x58e5,  0x6886,  0x78a7,  0x0840,  0x1861,  0x2802,  0x3823,
    0xc9cc,  0xd9ed,  0xe98e,  0xf9af,  0x8948,  0x9969,  0xa90a,  0xb92b,
    0x5af5,  0x4ad4,  0x7ab7,  0x6a96,  0x1a71,  0x0a50,  0x3a33,  0x2a12,
    0xdbfd,  0xcbdc,  0xfbbf,  0xeb9e,  0x9b79,  0x8b58,  0xbb3b,  0xab1a,
    0x6ca6,  0x7c87,  0x4ce4,  0x5cc5,  0x2c22,  0x3c03,  0x0c60,  0x1c41,
    0xedae,  0xfd8f,  0xcdec,  0xddcd,  0xad2a,  0xbd0b,  0x8d68,  0x9d49,
    0x7e97,  0x6eb6,  0x5ed5,  0x4ef4,  0x3e13,  0x2e32,  0x1e51,  0x0e70,
    0xff9f,  0xefbe,  0xdfdd,  0xcffc,  0xbf1b,  0xaf3a,  0x9f59,  0x8f78,
    0x9188,  0x81a9,  0xb1ca,  0xa1eb,  0xd10c,  0xc12d,  0xf14e,  0xe16f,
    0x1080,  0x00a1,  0x30c2,  0x20e3,  0x5004,  0x4025,  0x7046,  0x6067,
    0x83b9,  0x9398,  0xa3fb,  0xb3da,  0xc33d,  0xd31c,  0xe37f,  0xf35e,
    0x02b1,  0x1290,  0x22f3,  0x32d2,  0x4235,  0x5214,  0x6277,  0x7256,
    0xb5ea,  0xa5cb,  0x95a8,  0x8589,  0xf56e,  0xe54f,  0xd52c,  0xc50d,
    0x34e2,  0x24c3,  0x14a0,  0x0481,  0x7466,  0x6447,  0x5424,  0x4405,
    0xa7db,  0xb7fa,  0x8799,  0x97b8,  0xe75f,  0xf77e,  0xc71d,  0xd73c,
    0x26d3,  0x36f2,  0x0691,  0x16b0,  0x6657,  0x7676,  0x4615,  0x5634,
    0xd94c,  0xc96d,  0xf90e,  0xe92f,  0x99c8,  0x89e9,  0xb98a,  0xa9ab,
    0x5844,  0x4865,  0x7806,  0x6827,  0x18c0,  0x08e1,  0x3882,  0x28a3,
    0xcb7d,  0xdb5c,  0xeb3f,  0xfb1e,  0x8bf9,  0x9bd8,  0xabbb,  0xbb9a,
    0x4a75,  0x5a54,  0x6a37,  0x7a16,  0x0af1,  0x1ad0,  0x2ab3,  0x3a92,
    0xfd2e,  0xed0f,  0xdd6c,  0xcd4d,  0xbdaa,  0xad8b,  0x9de8,  0x8dc9,
    0x7c26,  0x6c07,  0x5c64,  0x4c45,  0x3ca2,  0x2c83,  0x1ce0,  0x0cc1,
    0xef1f,  0xff3e,  0xcf5d,  0xdf7c,  0xaf9b,  0xbfba,  0x8fd9,  0x9ff8,
    0x6e17,  0x7e36,  0x4e55,  0x5e74,  0x2e93,  0x3eb2,  0x0ed1,  0x1ef0
    };

WORD CalcCRC(WORD    icrc,
            BYTE   *icp,
            WORD    icnt)
	{
	register WORD crc = icrc;
	register BYTE *cp = icp;
	register WORD cnt = icnt;

	while( cnt-- )
		crc = (crc<<B) ^ crctab[(crc>>(W-B)) ^ *cp++];

	return( crc );
	}

